


EnableExplicit



Structure Drawing3D_Color
	Red.f
	Green.f
	Blue.f
	Alpha.f
EndStructure

Structure Drawing3D_Vector
	X.f
	Y.f
	Z.f
EndStructure

Structure Drawing3D_Matrix
	A11.f : A12.f : A13.f
	A21.f : A22.f : A23.f
	A31.f : A32.f : A33.f 
EndStructure

Structure Drawing3D_Pixel
	Color.Drawing3D_Color
	Distance.f
	*PreviousPixel.Drawing3D_Pixel
EndStructure

Structure Drawing3D_Triangle
	Vertex.Drawing3D_Vector[3]
	Color.Drawing3D_Color
EndStructure

Structure Drawing3D_Line
	Vertex.Drawing3D_Vector[2]
	Color.Drawing3D_Color
EndStructure

Structure Drawing3D_Light
	Direction.Drawing3D_Vector
	Color.Drawing3D_Color
EndStructure

Structure Drawing3D_Cluster
	Type.i
	Position.Drawing3D_Vector
	Orientation.Drawing3D_Matrix
	List		Triangle.Drawing3D_Triangle()
	List		*Cluster.Drawing3D_Cluster()
EndStructure

Structure Drawing3D_Object
	StructureUnion
		Type.i
		Cluster.Drawing3D_Cluster
		Line.Drawing3D_Line
		Triangle.Drawing3D_Triangle
	EndStructureUnion
EndStructure

Structure Drawing3DInclude
	Array		*PixelIndex.Drawing3D_Pixel(0, 0) ; Index der Pixel fr schnelleren Zugriff
	List		Pixel.Drawing3D_Pixel()           ; Liste aller zu zeichnenden Pixel
	List		Cluster.Drawing3D_Cluster()
	List		Light.Drawing3D_Light()           ; Liste aller Lichtquellen
	*CurrentCluster.Drawing3D_Cluster
	Orientation.Drawing3D_Matrix              ; Orientierung der Szene
	Position.Drawing3D_Vector                 ; Position der Kamera
	Background.Drawing3D_Color                ; Hintergrundfarbe
	MainCluster.Drawing3D_Cluster
	MaxX.i
	MaxY.i
	CenterX.i
	CenterY.i
	Distance.f
EndStructure


Global Drawing3DInclude.Drawing3DInclude





;- General

Procedure.f Drawing3D_Min(Value1.f, Value2.f, Value3.f=1e30)
	If Value3 < Value2
		If Value3 < Value1
			ProcedureReturn Value3
		Else
			ProcedureReturn Value1
		EndIf
	ElseIf Value2 < Value1
		ProcedureReturn Value2
	Else
		ProcedureReturn Value1
	EndIf
EndProcedure

Procedure.f Drawing3D_Max(Value1.f, Value2.f, Value3.f=-1e30)
	If Value3 > Value2
		If Value3 > Value1
			ProcedureReturn Value3
		Else
			ProcedureReturn Value1
		EndIf
	ElseIf Value2 > Value1
		ProcedureReturn Value2
	Else
		ProcedureReturn Value1
	EndIf
EndProcedure



;- Color

Procedure.i Drawing3D_SetColor(*Use.Drawing3D_Color, Color.l)
	Protected Factor.f = Alpha(Color) / 255 / 255
	*Use\Alpha = Alpha(Color) / 255
	*Use\Red   = Red(Color) * Factor
	*Use\Green = Green(Color) * Factor
	*Use\Blue  = Blue(Color) * Factor
	ProcedureReturn *Use
EndProcedure

Procedure.l Drawing3D_GetColor(*Source.Drawing3D_Color)
	Protected Factor.f, A, R, G, B
	If *Source\Alpha
		Factor = 255.0/*Source\Alpha
		A = *Source\Alpha*255
		R = *Source\Red*Factor
		G = *Source\Green*Factor
		B = *Source\Blue*Factor
		ProcedureReturn R|G<<8|B<<16|A<<24
	Else
		ProcedureReturn 0
	EndIf
EndProcedure

Procedure Drawing3D_AlphaBlend(*Use.Drawing3D_Color, *Source.Drawing3D_Color)
	Protected InvAlpha.f = 1.0 - *Source\Alpha
	*Use\Red   * InvAlpha + *Source\Red
	*Use\Green * InvAlpha + *Source\Green
	*Use\Blue  * InvAlpha + *Source\Blue
	*Use\Alpha * InvAlpha + *Source\Alpha
	ProcedureReturn *Use
EndProcedure

Procedure.i Drawing3D_NormalizeColor(*Use.Drawing3D_Color)
	Protected Max.f = Drawing3D_Max(*Use\Red, *Use\Green, *Use\Blue)
	If Max > 1
		*Use\Red / Max
		*Use\Green / Max
		*Use\Blue / Max
	EndIf
	*Use\Red * *Use\Alpha
	*Use\Green * *Use\Alpha
	*Use\Blue * *Use\Alpha
	ProcedureReturn *Use
EndProcedure




;- Math

Procedure.i Drawing3D_Vector(*Use.Drawing3D_Vector, X.f, Y.f, Z.f)
	*Use\X = X
	*Use\Y = Y
	*Use\Z = Z
	ProcedureReturn *Use
EndProcedure

Procedure.f Drawing3D_Length(*Source.Drawing3D_Vector)
	ProcedureReturn Sqr(*Source\X * *Source\X + *Source\Y * *Source\Y + *Source\Z * *Source\Z)
EndProcedure

Procedure.i Drawing3D_Normalize(*Use.Drawing3D_Vector)
	Protected Length.f = Drawing3D_Length(*Use)
	If Length
		*Use\X / Length
		*Use\Y / Length
		*Use\Z / Length
	EndIf
	ProcedureReturn *Use
EndProcedure

Procedure.i Drawing3D_Add(*Use.Drawing3D_Vector, *Source.Drawing3D_Vector)
	*Use\X + *Source\X
	*Use\Y + *Source\Y
	*Use\Z + *Source\Z
	ProcedureReturn *Use
EndProcedure

Procedure.i Drawing3D_Subtract(*Use.Drawing3D_Vector, *Source.Drawing3D_Vector)
	*Use\X - *Source\X
	*Use\Y - *Source\Y
	*Use\Z - *Source\Z
	ProcedureReturn *Use
EndProcedure

Procedure.i Drawing3D_Cross(*Use.Drawing3D_Vector, *Source1.Drawing3D_Vector, *Source2.Drawing3D_Vector)
	*Use\X = *Source1\Y * *Source2\Z - *Source1\Z * *Source2\Y
	*Use\Y = *Source1\Z * *Source2\X - *Source1\X * *Source2\Z
	*Use\Z = *Source1\X * *Source2\Y - *Source1\Y * *Source2\X
	ProcedureReturn *Use
EndProcedure

Procedure.f Drawing3D_Scalar(*Source1.Drawing3D_Vector, *Source2.Drawing3D_Vector)
	ProcedureReturn *Source1\X * *Source2\X + *Source1\Y * *Source2\Y + *Source1\Z * *Source2\Z
EndProcedure

Procedure.i Drawing3D_Rotate(*Use.Drawing3D_Vector, *Source.Drawing3D_Matrix)
	Protected Buffer.Drawing3D_Vector 
	CopyMemory(*Use, @Buffer, SizeOf(Drawing3D_Vector))
	*Use\X = Buffer\X**Source\A11 + Buffer\Y**Source\A12 + Buffer\Z**Source\A13
	*Use\Y = Buffer\X**Source\A21 + Buffer\Y**Source\A22 + Buffer\Z**Source\A23
	*Use\Z = Buffer\X**Source\A31 + Buffer\Y**Source\A32 + Buffer\Z**Source\A33
	ProcedureReturn *Use
EndProcedure

Procedure.i Drawing3D_Orientation(*Use.Drawing3D_Matrix, X.f, Y.f, Z.f)
	Protected CosZ.f = Cos(Z), CosY.f = Cos(Y), CosX.f = Cos(X)
	Protected SinZ.f = Sin(Z), SinY.f = Sin(Y), SinX.f = Sin(X)
	*Use\A11 =  CosY*CosZ                : *Use\A12 = -CosY*SinZ                : *Use\A13 =  SinY
	*Use\A21 =  SinX*SinY*CosZ+CosX*SinZ : *Use\A22 = -SinX*SinY*SinZ+CosX*CosZ : *Use\A23 = -SinX*CosY
	*Use\A31 = -CosX*SinY*CosZ+SinX*SinZ : *Use\A32 =  CosX*SinY*SinZ+SinX*CosZ : *Use\A33 =  CosX*CosY
	ProcedureReturn *Use
EndProcedure

Procedure Drawing3D_Multiply(*Use.Drawing3D_Matrix, *Source.Drawing3D_Matrix)
	Protected Buffer.Drawing3D_Matrix
	CopyMemory(*Use, @Buffer, SizeOf(Drawing3D_Matrix)) 
	With *Use
		\A11 = Buffer\A11 * *Source\A11 + Buffer\A12 * *Source\A21 + Buffer\A13 * *Source\A31
		\A12 = Buffer\A11 * *Source\A12 + Buffer\A12 * *Source\A22 + Buffer\A13 * *Source\A32
		\A13 = Buffer\A11 * *Source\A13 + Buffer\A12 * *Source\A23 + Buffer\A13 * *Source\A33
		\A21 = Buffer\A21 * *Source\A11 + Buffer\A22 * *Source\A21 + Buffer\A23 * *Source\A31
		\A22 = Buffer\A21 * *Source\A12 + Buffer\A22 * *Source\A22 + Buffer\A23 * *Source\A32
		\A23 = Buffer\A21 * *Source\A13 + Buffer\A22 * *Source\A23 + Buffer\A23 * *Source\A33
		\A31 = Buffer\A31 * *Source\A11 + Buffer\A32 * *Source\A21 + Buffer\A33 * *Source\A31
		\A32 = Buffer\A31 * *Source\A12 + Buffer\A32 * *Source\A22 + Buffer\A33 * *Source\A32
		\A33 = Buffer\A31 * *Source\A13 + Buffer\A32 * *Source\A23 + Buffer\A33 * *Source\A33
	EndWith
	ProcedureReturn *Use
EndProcedure




;- Projection

Procedure.i Drawing3D_Project(*Use.Drawing3D_Vector)
	If *Use\Z
		*Use\X = - *Use\X * Drawing3DInclude\Distance / *Use\Z + Drawing3DInclude\CenterX
		*Use\Y =   *Use\Y * Drawing3DInclude\Distance / *Use\Z + Drawing3DInclude\CenterY
	EndIf
	ProcedureReturn *Use
EndProcedure

Procedure.i Drawing3D_Projection(*Use.Drawing3D_Vector, *Source.Drawing3D_Vector)
	If *Source\Z
		*Use\X = - *Source\X * Drawing3DInclude\Distance / *Source\Z + Drawing3DInclude\CenterX
		*Use\Y =   *Source\Y * Drawing3DInclude\Distance / *Source\Z + Drawing3DInclude\CenterY
	EndIf
	ProcedureReturn *Use
EndProcedure



;- Cluster (unfertig)

Procedure.i CreateCluster3D()
	Protected *Cluster.Drawing3D_Cluster = AddElement(Drawing3DInclude\Cluster())
	Drawing3DInclude\CurrentCluster = *Cluster
	ProcedureReturn *Cluster
EndProcedure

Procedure CloseCluster3D()
EndProcedure

Procedure Cluster3D(*Cluster.Drawing3D_Cluster, X.f, Y.f, Z.f, Pitch.f=0.0, Yaw.f=0.0, Roll.f=0.0)
EndProcedure



;- Drawing

Procedure Light3D(DirectionX.f, DirectionY.f, DirectionZ.f, Color.l)
	
	Protected *Light.Drawing3D_Light = AddElement(Drawing3DInclude\Light())
	
	With *Light
		Drawing3D_Vector(\Direction, DirectionX, DirectionY, DirectionZ)
		Drawing3D_Normalize(\Direction)
		Drawing3D_SetColor(\Color, Color)
	EndWith
	
	ProcedureReturn *Light
	
EndProcedure

Procedure Drawing3D_DrawTriangle(*Triangle.Drawing3D_Triangle)
	
	Protected.Drawing3D_Vector Point0, Point1, Point2
	Protected LightFactor.f
	
	With *Triangle
		
		Drawing3D_Projection(Point0, \Vertex[0])
		Drawing3D_Projection(Point1, \Vertex[1])
		Drawing3D_Projection(Point2, \Vertex[2])
		
		Protected X.i, Y.i
		Protected DX21.f = Point2\X-Point1\X, DY21.f = Point2\Y-Point1\Y
		Protected DX02.f = Point0\X-Point2\X, DY02.f = Point0\Y-Point2\Y
		Protected DX10.f = Point1\X-Point0\X, DY10.f = Point1\Y-Point0\Y
		Protected L1.f, L2.f, L3.f
		
		Protected XA.i = Drawing3D_Max(Drawing3D_Min(Point0\X, Point1\X, Point2\X), 0)
		Protected XB.i = Drawing3D_Min(Drawing3D_Max(Point0\X, Point1\X, Point2\X), Drawing3DInclude\MaxX)
		Protected YA.i = Drawing3D_Max(Drawing3D_Min(Point0\Y, Point1\Y, Point2\Y), 0)
		Protected YB.i = Drawing3D_Min(Drawing3D_Max(Point0\Y, Point1\Y, Point2\Y), Drawing3DInclude\MaxY)
		
		Protected *Pixel.Drawing3D_Pixel, *OldPixel.Drawing3D_Pixel, Color.Drawing3D_Color, Norm.Drawing3D_Vector, V1.Drawing3D_Vector, V2.Drawing3D_Vector
		
		Protected T.f = 1.0 / (-DY21*DX02 + DX21*DY02)
		Protected Distance.f, Brightness.f, Q1.f, Q2.f
		
		;If T< 0 : ProcedureReturn 0 : EndIf
		
		Color = \Color
		Color\Red / \Color\Alpha
		Color\Green / \Color\Alpha
		Color\Blue / \Color\Alpha
		V1 = \Vertex[1] : Drawing3D_Subtract(V1, \Vertex[0])
		V2 = \Vertex[2] : Drawing3D_Subtract(V2, \Vertex[0])
		Drawing3D_Normalize(Drawing3D_Cross(Norm, V2, V1))
		ForEach Drawing3DInclude\Light()
			LightFactor = Drawing3D_Scalar(Norm, Drawing3DInclude\Light()\Direction)*Sign(T)
			If LightFactor > 0
				Color\Red   + LightFactor*Drawing3DInclude\Light()\Color\Red
				Color\Green + LightFactor*Drawing3DInclude\Light()\Color\Green
				Color\Blue  + LightFactor*Drawing3DInclude\Light()\Color\Blue
			EndIf
		Next
		Drawing3D_NormalizeColor(Color)
		
		
		For Y = YA To YB
			Q1 = DX21*(Y-Point2\Y)
			Q2 = DX02*(Y-Point2\Y)
			*Pixel = #Null
			For X = XA To XB
				L1 = ( -DY21*(X-Point2\X) + Q1 ) * T
				L2 = ( -DY02*(X-Point2\X) + Q2 ) * T
				L3 = 1.0 - L1 - L2
				If L1 >= 0.0 And L2 >= 0.0 And L3 >= 0.0
					*Pixel = AddElement(Drawing3DInclude\Pixel())
					*OldPixel = Drawing3DInclude\PixelIndex(X, Y)
					Distance = L1/\Vertex[0]\Z + L2/\Vertex[1]\Z + L3/\Vertex[2]\Z
					If Distance > 0
						Continue
					EndIf
					If *OldPixel
						If Distance < *OldPixel\Distance
							*Pixel\PreviousPixel = *OldPixel
							Drawing3DInclude\PixelIndex(X, Y) = *Pixel
						Else
							While *OldPixel\PreviousPixel And *OldPixel\PreviousPixel\Distance < Distance
								*OldPixel = *OldPixel\PreviousPixel
							Wend
							*Pixel\PreviousPixel = *OldPixel\PreviousPixel
							*OldPixel\PreviousPixel = *Pixel
						EndIf
					Else
						Drawing3DInclude\PixelIndex(X, Y) = *Pixel
					EndIf
					*Pixel\Distance    = Distance
					*Pixel\Color\Red   = Color\Red
					*Pixel\Color\Green = Color\Green
					*Pixel\Color\Blue  = Color\Blue
					*Pixel\Color\Alpha = Color\Alpha
				ElseIf *Pixel
					Break
				EndIf
			Next
		Next
		
	EndWith
	
EndProcedure

Procedure Drawing3D_DrawLine(*Line.Drawing3D_Line)
	
	Protected.Drawing3D_Vector Point0, Point1
	Protected X.f, Y.f, I.f
	
	With *Line
		
		Drawing3D_Projection(Point0, \Vertex[0])
		Drawing3D_Projection(Point1, \Vertex[1])
		
		Protected Length.f = Sqr(Point0\X*Point0\X+Point0\Y*Point0\Y)
		Protected Distance.f
		
		Protected DX.f = (Point1\X-Point0\X)/Length/2
		Protected DY.f = (Point1\Y-Point0\Y)/Length/2
		
		Protected *Pixel.Drawing3D_Pixel, *OldPixel.Drawing3D_Pixel, Color.Drawing3D_Color, Norm.Drawing3D_Vector, V1.Drawing3D_Vector, V2.Drawing3D_Vector
		
		X = Point0\X
		Y = Point0\Y
		While I < Length
			If X >= 1 And Y >= 1 And X <= Drawing3DInclude\MaxX-1 And Y <= Drawing3DInclude\MaxY-1
				*Pixel = AddElement(Drawing3DInclude\Pixel())
				*OldPixel = Drawing3DInclude\PixelIndex(Int(X), Int(Y))
				Distance = (1-I/Length)/\Vertex[0]\Z + (I/Length)/\Vertex[1]\Z
				If Distance > 0
					Continue
				EndIf
				If *OldPixel
					If Distance < *OldPixel\Distance
						*Pixel\PreviousPixel = *OldPixel
						Drawing3DInclude\PixelIndex(Int(X), Int(Y)) = *Pixel
					Else
						While *OldPixel\PreviousPixel And *OldPixel\PreviousPixel\Distance < Distance
							*OldPixel = *OldPixel\PreviousPixel
						Wend
						*Pixel\PreviousPixel = *OldPixel\PreviousPixel
						*OldPixel\PreviousPixel = *Pixel
					EndIf
				Else
					Drawing3DInclude\PixelIndex(Int(X), Int(Y)) = *Pixel
				EndIf
				*Pixel\Distance    = Distance
				*Pixel\Color\Red   = \Color\Red
				*Pixel\Color\Green = \Color\Green
				*Pixel\Color\Blue  = \Color\Blue
				*Pixel\Color\Alpha = \Color\Alpha
			EndIf
			X + DX
			Y + DY
			I + 0.5
		Wend
		
	EndWith
	
EndProcedure




Procedure StartDrawing3D(Output.i, FieldOfView.f=75)
	With Drawing3DInclude
		If StartDrawing(Output)
			\MaxX = OutputWidth()-1
			\MaxY = OutputHeight()-1
			\CenterX = OutputWidth()/2
			\CenterY = OutputHeight()/2
			\Distance = OutputHeight()/Tan(Radian(FieldOfView)*0.5)
			If ArraySize(\PixelIndex(), 1) <> \MaxX Or ArraySize(\PixelIndex(), 2) <> \MaxY 
				Dim \PixelIndex(\MaxX, \MaxY)
			Else
				FillMemory(@\PixelIndex(0,0), SizeOf(Integer)*(\MaxX+1)*(\MaxY+1))
			EndIf
			ClearList(\Pixel())
			ClearList(\Light())
			ClearStructure(\MainCluster, Drawing3D_Cluster)
			ProcedureReturn #True
		EndIf
	EndWith
EndProcedure

Procedure StopDrawing3D()
	Protected MaxX = ArraySize(Drawing3DInclude\PixelIndex(), 1)
	Protected MaxY = ArraySize(Drawing3DInclude\PixelIndex(), 2)
	Protected *Pixel.Drawing3D_Pixel, *PreviousPixel.Drawing3D_Pixel, Color.Drawing3D_Color
	Protected X, Y
	For Y = 0 To Drawing3DInclude\MaxY
		For X = 0 To Drawing3DInclude\MaxX
			*Pixel = Drawing3DInclude\PixelIndex(X, Y)
			If *Pixel
				While *Pixel\PreviousPixel
					Drawing3D_AlphaBlend(*Pixel\PreviousPixel\Color, *Pixel\Color)
					*Pixel = *Pixel\PreviousPixel
				Wend
				Color = Drawing3DInclude\Background
				Plot(X, Y, Drawing3D_GetColor(Drawing3D_AlphaBlend(Color, *Pixel))&$FFFFFF)
			EndIf
		Next
	Next
	StopDrawing()
EndProcedure

Procedure Clear3D(Color.l=$FF000000)
	Box(0, 0, OutputWidth(), OutputHeight(), Color)
	Drawing3D_SetColor(Drawing3DInclude\Background, Color)
EndProcedure

Procedure Line3D(X1.f, Y1.f, Z1.f, X2.f, Y2.f, Z2.f, Color.l)
	
	Protected Line.Drawing3D_Line
	Protected N.i
	
	Drawing3D_Vector(Line\Vertex[0], X1, Y1, Z1)
	Drawing3D_Vector(Line\Vertex[1], X2, Y2, Z2)
	Drawing3D_SetColor(Line\Color, Color)
	For N = 0 To 1
		Drawing3D_Rotate(Line\Vertex[N], Drawing3DInclude\Orientation)
		Drawing3D_Subtract(Line\Vertex[N], Drawing3DInclude\Position)
	Next
	Drawing3D_DrawLine(Line)
	
EndProcedure



Procedure Triangle3D(X1.f, Y1.f, Z1.f, X2.f, Y2.f, Z2.f, X3.f, Y3.f, Z3.f, Color.l)
	
	Protected Triangle.Drawing3D_Triangle
	Protected N.i
	
	Drawing3D_Vector(Triangle\Vertex[0], X1, Y1, Z1)
	Drawing3D_Vector(Triangle\Vertex[1], X2, Y2, Z2)
	Drawing3D_Vector(Triangle\Vertex[2], X3, Y3, Z3)
	Drawing3D_SetColor(Triangle\Color, Color)
	For N = 0 To 2
		Drawing3D_Rotate(Triangle\Vertex[N], Drawing3DInclude\Orientation)
		Drawing3D_Subtract(Triangle\Vertex[N], Drawing3DInclude\Position)
	Next
	Drawing3D_DrawTriangle(Triangle)
	
EndProcedure



Procedure Plane3D(X1.f, Y1.f, Z1.f, X2.f, Y2.f, Z2.f, X3.f, Y3.f, Z3.f, X4.f, Y4.f, Z4.f, Color.l)
	
	Protected X.f = (X1+X2+X3+X4)*0.25
	Protected Y.f = (Y1+Y2+Y3+Y4)*0.25
	Protected Z.f = (Z1+Z2+Z3+Z4)*0.25
	
	;Triangle3D(X1, Y1, Z1, X2, Y2, Z2, X, Y, Z, Color)
	;Triangle3D(X2, Y2, Z2, X4, Y4, Z4, X, Y, Z, Color)
	;Triangle3D(X4, Y4, Z4, X3, Y3, Z3, X, Y, Z, Color)
	;Triangle3D(X3, Y3, Z3, X1, Y1, Z1, X, Y, Z, Color)
	
	Triangle3D(X1, Y1, Z1, X2, Y2, Z2, X3, Y3, Z3, Color)
	Triangle3D(X2, Y2, Z2, X4, Y4, Z4, X3, Y3, Z3, Color)
	
EndProcedure


Drawing3D_Orientation(Drawing3DInclude\Orientation, 0, 0, 0)
; IDE Options = PureBasic 5.00 Beta 3 (Windows - x86)
; CursorPosition = 497
; FirstLine = 22
; Folding = ---------
; EnableXP
; DisableDebugger
; EnableCompileCount = 536
; EnableBuildCount = 0